package require tdom

namespace eval kp {
  variable nodeId [clock milliseconds][expr round(rand()*1000)]
  variable transactionId 1
  
  # namespaces
  array set ns {
   rdf http://www.w3.org/1999/02/22-rdf-syntax-ns#
   tue http://www.win.tue.nl/~mholende/m3#
  }
}

proc kp::extractValue {root what} {
  set node [$root selectNodes .//$what]
  set value [[$node childNodes] nodeValue]
  $node delete
  return $value
}

proc kp::extractTriple {node} {
  set s [kp::extractValue $node subject]
  set p [kp::extractValue $node predicate]
  set o [kp::extractValue $node object]
  
  # substitute namespace short names
  foreach element {s o p} {
    foreach {short long} [array get kp::ns] {
      regsub $long [set $element] ${short}: $element
    }
  }
  
  return [list $s $p $o]
}

proc kp::extractTriples {root args} {
  if {[llength $args]} {
    set args {*}$args
    set root [$root selectNodes .//$args]
  }
  
  set triples {}
  set nodes [$root selectNodes .//triple_list/*]
  foreach node $nodes {
    lappend triples {*}[kp::extractTriple $node]
  }
  return $triples
}

proc kp::extractUris {root} {
  set uris {}
  set nodes [$root selectNodes .//urilist/*]
  foreach node $nodes {
    lappend uris [[$node childNodes] nodeValue]
  }
  return $uris
}

set kp::triple {
  <triple>
    <subject type="{subjectType}">{subject}</subject>
    <predicate>{predicate}</predicate>
    <object type="{objectType}">{object}</object>
  </triple>
}

#
# subjectType: uri | bnode
# objectType:  uri | bnode | literal
#

proc kp::triple {subject predicate object subjectType objectType} {
  foreach field {subject predicate object} {
    if {"*" eq [set $field]} {
      set $field sib:any
    }
  }

  set xml $kp::triple
  foreach field {subject predicate object subjectType objectType} {
    regsub -all "\{$field\}" $xml [set $field] xml
  }
  return $xml
}

proc kp::incoming {sock} {
  variable incomingXML
  
  append incomingXML [read $sock]
  
  # extract the first complete message
  while {[regexp {<SSAP_message>.*?</SSAP_message>} $incomingXML message]} {
    regsub {<SSAP_message>.*?</SSAP_message>} $incomingXML "" incomingXML
    #puts $message
        
    set doc [dom parse $message]

    set type [kp::extractValue $doc transaction_type]
    set type [string tolower $type]
  
    set messageType [kp::extractValue $doc message_type]
    set messageType [string toupper [string tolower $messageType] 0]
  
    switch -- $messageType {
      Confirm {
        set status [kp::extractValue $doc {parameter[@name='status']}]  
      } 
      default {
        # if there was no status parameter, then assume success
        set status "m3:Success"
      }
    }
  
    if {![string match "m3:Success" $status]} {
      set messageType Error    
      puts "something went wrong on the SIB"
    }

    # call the handler, if specified
    set method $type$messageType
    if {[llength [info procs $method]]} {
      eval $method $doc 
    } else {
      puts "method $method not implemented"
    }
  
    $doc delete
  
    set ::incoming 1
  }
}

proc kp::send {type fields args} {
  # accept a dedicated socket for communication (e.g. for subscriptions)
  if {[llength $args]} {
    set sock $args
  } else {
    variable sock
  }
  
  if {![info exists sock]} { error "uninitialized socket" }

  incr kp::transactionId
  
  # add the standard fields
  lappend fields nodeId $kp::nodeId
  lappend fields spaceId $kp::spaceId
  lappend fields transactionId $kp::transactionId

  # instantiate the message template
  variable $type
  set xml [set $type]  
  foreach {k v} $fields {
    regsub -all "\{$k\}" $xml $v xml
  }

  # send the message
  puts $sock $xml
  flush $sock
  #puts send:$xml
  
  return kp::trans$kp::transactionId
}

proc kp::trigger {doc value} {
  variable transId [kp::extractValue $doc transaction_id]
  set kp::trans$transId $value
}

proc kp::wait {var} {
  vwait $var
  set result [set $var]
  unset $var
  return $result
}

proc kp::waitAny {var} {
  vwait $var
  set result [expr ![llength $var]]
  unset $var
  return $result
}

#
# JOIN
#

set kp::join {
  <SSAP_message>
    <transaction_type>JOIN</transaction_type>
    <message_type>REQUEST</message_type>
    <transaction_id>{transactionId}</transaction_id>
    <node_id>{nodeId}</node_id>
    <space_id>{spaceId}</space_id>
    <parameter name = "credentials">XYZZY</parameter>
  </SSAP_message>
}

proc kp::join {_ip _port id} {
  # setup the socket
  variable sock [socket $_ip $_port]
  chan configure $sock -blocking 0 -translation binary
  chan event $sock readable [list kp::incoming $sock]

  # send the join message
  variable spaceId $id
  set var [send join {}]

  variable ip $_ip
  variable port $_port
  
  return [kp::waitAny $var]
}

proc kp::joinConfirm {doc} {
  # store the spaceId and nodeId for future reference
  variable spaceId [kp::extractValue $doc space_id]
  variable nodeId [kp::extractValue $doc node_id]
  
  kp::trigger $doc ok
}

proc kp::joinError {doc} {
  # store the spaceId and nodeId for future reference
  variable spaceId
  unset spaceId

  kp::trigger $doc error
}

#
# LEAVE
#

set kp::leave {
  <SSAP_message>
    <transaction_type>LEAVE</transaction_type>
    <message_type>REQUEST</message_type>
    <transaction_id>{transactionId}</transaction_id>
    <node_id>{nodeId}</node_id>
    <space_id>{spaceId}</space_id>
  </SSAP_message>
}

proc kp::leave {} {
  set var [send leave {}]
  return [kp::waitAny $var]
}

proc kp::leaveConfirm {doc} {
  variable sock
  close $sock
  unset sock
  
  kp::trigger $doc ""
}

#
# QUERY
#

set kp::query {
  <SSAP_message>
    <transaction_type>QUERY</transaction_type>
    <message_type>REQUEST</message_type>
    <transaction_id>{transactionId}</transaction_id>
    <node_id>{nodeId}</node_id>
    <space_id>{spaceId}</space_id>
    <parameter name="type">RDF-M3</parameter>
    <parameter name="query">
      <triple_list>
        {triples}
      </triple_list>
    </parameter>
  </SSAP_message>
}

proc kp::query {subject predicate object args} {
  set subjectType uri
  set objectType uri
  foreach {subjectType objectType} $args break

  set triples [kp::triple $subject $predicate $object $subjectType $objectType]
  set var [send query [list triples $triples]]
  
  return [kp::wait $var]
}

proc kp::queryConfirm {doc} {
  kp::trigger $doc [kp::extractTriples $doc]
}

# Utility method, returns an empty string of object not found

proc kp::queryObject {subject predicate} {
  set object ""
  foreach {_ _ object} [kp::query $subject $predicate *] break
  if {![llength $object]} {
    puts "no object satisfies the predicate '$subject $predicate *'"
  }
  return $object
}

#
# INSERT
#

set kp::insert {
  <SSAP_message>
    <transaction_type>INSERT</transaction_type>
    <message_type>REQUEST</message_type>
    <transaction_id>{transactionId}</transaction_id>
    <node_id>{nodeId}</node_id>
    <space_id>{spaceId}</space_id>
    <parameter name="insert_graph" encoding="RDF-M3">
      <triple_list>
        {triples}
      </triple_list>
    </parameter>
    <parameter name = "confirm">TRUE</parameter>
  </SSAP_message>
}

proc kp::insert {subject predicate object args} {
  set subjectType uri
  set objectType uri
  foreach {subjectType objectType} $args break
  
  set triples [kp::triple $subject $predicate $object $subjectType $objectType]
  set var [send insert [list triples $triples]]

  return [kp::wait $var]
}

proc kp::insertConfirm {doc} {
  kp::trigger $doc [extractUris $doc]
}

#
# REMOVE
#

set kp::remove {
  <SSAP_message>
    <transaction_type>REMOVE</transaction_type>
    <message_type>REQUEST</message_type>
    <transaction_id>{transactionId}</transaction_id>
    <node_id>{nodeId}</node_id>
    <space_id>{spaceId}</space_id>
    <parameter name="remove_graph" encoding="RDF-M3">
      <triple_list>
        {triples}
      </triple_list>
    </parameter>
    <parameter name = "confirm">TRUE</parameter>
  </SSAP_message>
}

proc kp::remove {subject predicate object args} {
  set subjectType uri
  set objectType uri
  foreach {subjectType objectType} $args break

  set triples [kp::triple $subject $predicate $object $subjectType $objectType]
  set var [send remove [list triples $triples]]

  return [kp::waitAny $var]
}

set kp::removeConfirmUri {
  <SSAP_message>
      <transaction_id>5</transaction_id>
      <space_id>X</space_id>
      <node_id>1379669343635</node_id>
  </SSAP_message>
}

proc kp::removeConfirm {doc} {
  kp::trigger $doc ""
}

#
# UPDATE
#

set kp::update {
  <SSAP_message>
    <transaction_type>UPDATE</transaction_type>
    <message_type>REQUEST</message_type>
    <transaction_id>{transactionId}</transaction_id>
    <node_id>{nodeId}</node_id>
    <space_id>{spaceId}</space_id>
    <parameter name="remove_graph" encoding="RDF-M3">
      <triple_list>
        {removeTriples}
      </triple_list>
    </parameter>
    <parameter name="insert_graph" encoding="RDF-M3">
      <triple_list>
        {insertTriples}
      </triple_list>
    </parameter>
    <parameter name = "confirm">TRUE</parameter>
  </SSAP_message>
}

proc kp::update {subject predicate object args} {
  set subjectType uri
  set objectType uri
  foreach {subjectType objectType} $args break

  set removeTriples [kp::triple $subject $predicate * $subjectType uri]
  set insertTriples [kp::triple $subject $predicate $object $subjectType $objectType]  
  
  set var [send update [list removeTriples $removeTriples insertTriples $insertTriples]]

  return [kp::waitAny $var]
}

proc kp::updateConfirm {doc} {
  kp::trigger $doc ""
}

#
# SUBSCRIBE
#

set kp::subscribe {
  <SSAP_message>
    <transaction_type>SUBSCRIBE</transaction_type>
    <message_type>REQUEST</message_type>
    <transaction_id>{transactionId}</transaction_id>
    <node_id>{nodeId}</node_id>
    <space_id>{spaceId}</space_id>
    <parameter name="type">RDF-M3</parameter>
    <parameter name="query">
      <triple_list>
        {triples}
      </triple_list>
    </parameter>
  </SSAP_message>
}

proc kp::subscribe {handler subject predicate object args} {
  #puts "kp::subscribe $handler $subject $predicate $object"
  set subjectType uri
  set objectType uri
  foreach {subjectType objectType} $args break

  # open a new socket for the subscription
  variable ip
  variable port
  set sock [socket $ip $port]
  chan configure $sock -blocking 0 -translation binary
  chan event $sock readable [list kp::incoming $sock]
  
  # register the subscription at the SIB
  set triples [kp::triple $subject $predicate $object $subjectType $objectType]
  set var [send subscribe [list triples $triples] $sock]
  
  # wait for the subscription to be confirmed
  set s [kp::wait $var]

  # bind a handler to the subscription
  array set sub $s
  variable subscriptionHandler
  set subscriptionHandler($sub(id)) $handler
  
  # handle any new indications piggybacked onto the confirmation message
  $handler $sub(triples) {}
  
  return $sub(id)
}

proc kp::subscribeConfirm {doc} {
  set id [extractValue $doc {parameter[@name='subscription_id']}]
  set triples [extractTriples $doc]

  kp::trigger $doc [list id $id triples $triples]
}

proc kp::subscribeIndication {doc} {
  set id [extractValue $doc {parameter[@name='subscription_id']}]

  # call the handler
  variable subscriptionHandler
  set new [extractTriples $doc {parameter[@name="new_results"]}]
  set obsolete [extractTriples $doc {parameter[@name="obsolete_results"]}]
  $subscriptionHandler($id) $new $obsolete
}
